######################################################################
#
# File: test/integration/helpers.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import dataclasses
import json
import logging
import os
import pathlib
import platform
import random
import re
import secrets
import shutil
import string
import subprocess
import sys
import threading
import time
import warnings
from dataclasses import dataclass
from datetime import datetime, timedelta
from os import environ, linesep
from pathlib import Path
from tempfile import mkdtemp, mktemp

import backoff
from b2sdk.v2 import (
    ALL_CAPABILITIES,
    BUCKET_NAME_CHARS_UNIQ,
    BUCKET_NAME_LENGTH_RANGE,
    NO_RETENTION_FILE_SETTING,
    B2Api,
    Bucket,
    EncryptionAlgorithm,
    EncryptionKey,
    EncryptionMode,
    EncryptionSetting,
    InMemoryAccountInfo,
    InMemoryCache,
    LegalHold,
    RetentionMode,
    fix_windows_path_limit,
)
from b2sdk.v2.exception import (
    BadRequest,
    BucketIdNotFound,
    FileNotPresent,
    TooManyRequests,
    v3BucketIdNotFound,
)

from b2._internal.console_tool import Command, current_time_millis

logger = logging.getLogger(__name__)

# A large period is set here to avoid issues related to clock skew or other time-related issues under CI
BUCKET_CLEANUP_PERIOD_MILLIS = timedelta(days=1).total_seconds() * 1000
ONE_HOUR_MILLIS = 60 * 60 * 1000
ONE_DAY_MILLIS = ONE_HOUR_MILLIS * 24

BUCKET_NAME_LENGTH = BUCKET_NAME_LENGTH_RANGE[1]
BUCKET_CREATED_AT_MILLIS = 'created_at_millis'

NODE_DESCRIPTION = f"{platform.node()}: {platform.platform()}"


def get_seed():
    """
    Get seed for random number generator.

    GH Actions machines seem to offer a very limited entropy pool
    """
    return b''.join(
        (
            secrets.token_bytes(32),
            str(time.time_ns()).encode(),
            NODE_DESCRIPTION.encode(),
            str(os.getpid()).encode(),  # needed due to pytest-xdist
            str(environ).encode('utf8', errors='ignore'
                               ),  # especially helpful under GitHub (and similar) CI
        )
    )


RNG = random.Random(get_seed())
RNG_SEED = RNG.randint(0, 2 << 31)
RNG_COUNTER = 0

if sys.version_info < (3, 9):
    RNG.randbytes = lambda n: RNG.getrandbits(n * 8).to_bytes(n, 'little')

SSE_NONE = EncryptionSetting(mode=EncryptionMode.NONE,)
SSE_B2_AES = EncryptionSetting(
    mode=EncryptionMode.SSE_B2,
    algorithm=EncryptionAlgorithm.AES256,
)
_SSE_KEY = RNG.randbytes(32)
SSE_C_AES = EncryptionSetting(
    mode=EncryptionMode.SSE_C,
    algorithm=EncryptionAlgorithm.AES256,
    key=EncryptionKey(secret=_SSE_KEY, key_id='user-generated-key-id')
)
SSE_C_AES_2 = EncryptionSetting(
    mode=EncryptionMode.SSE_C,
    algorithm=EncryptionAlgorithm.AES256,
    key=EncryptionKey(secret=_SSE_KEY, key_id='another-user-generated-key-id')
)


def random_token(length: int, chars=string.ascii_letters) -> str:
    return ''.join(RNG.choice(chars) for _ in range(length))


def bucket_name_part(length: int) -> str:
    assert length >= 1
    global RNG_COUNTER
    RNG_COUNTER += 1
    name_part = random_token(length, BUCKET_NAME_CHARS_UNIQ)
    logger.info('RNG_SEED: %s', RNG_SEED)
    logger.info('RNG_COUNTER: %i, length: %i', RNG_COUNTER, length)
    logger.info('name_part: %s', name_part)
    return name_part


@dataclass
class Api:
    account_id: str
    application_key: str
    realm: str
    general_bucket_name_prefix: str
    this_run_bucket_name_prefix: str

    api: B2Api = None
    bucket_name_log: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        info = InMemoryAccountInfo()
        cache = InMemoryCache()
        self.api = B2Api(info, cache=cache)
        self.api.authorize_account(self.realm, self.account_id, self.application_key)
        assert BUCKET_NAME_LENGTH - len(
            self.this_run_bucket_name_prefix
        ) > 5, self.this_run_bucket_name_prefix

    def new_bucket_name(self) -> str:
        bucket_name = self.this_run_bucket_name_prefix + bucket_name_part(
            BUCKET_NAME_LENGTH - len(self.this_run_bucket_name_prefix)
        )
        self.bucket_name_log.append(bucket_name)
        return bucket_name

    def new_bucket_info(self) -> dict:
        return {
            BUCKET_CREATED_AT_MILLIS: str(current_time_millis()),
            "created_by": NODE_DESCRIPTION,
        }

    def create_bucket(self, bucket_type: str = 'allPublic', **kwargs) -> Bucket:
        bucket_name = self.new_bucket_name()
        return self.api.create_bucket(
            bucket_name,
            bucket_type=bucket_type,
            bucket_info=self.new_bucket_info(),
            **kwargs,
        )

    def _should_remove_bucket(self, bucket: Bucket) -> tuple[bool, str]:
        if bucket.name.startswith(self.this_run_bucket_name_prefix):
            return True, 'it is a bucket for this very run'
        if bucket.name.startswith(self.general_bucket_name_prefix):
            if BUCKET_CREATED_AT_MILLIS in bucket.bucket_info:
                delete_older_than = current_time_millis() - BUCKET_CLEANUP_PERIOD_MILLIS
                this_bucket_creation_time = int(bucket.bucket_info[BUCKET_CREATED_AT_MILLIS])
                if this_bucket_creation_time < delete_older_than:
                    return True, f"this_bucket_creation_time={this_bucket_creation_time} < delete_older_than={delete_older_than}"
                return False, f"this_bucket_creation_time={this_bucket_creation_time} >= delete_older_than={delete_older_than}"
            else:
                return True, 'undefined ' + BUCKET_CREATED_AT_MILLIS
        return False, f'does not start with {self.general_bucket_name_prefix!r}'

    def clean_buckets(self, quick=False):
        # even with use_cache=True, if cache is empty API call will be made
        buckets = self.api.list_buckets(use_cache=quick)
        print('Total bucket count:', len(buckets))
        remaining_buckets = []
        for bucket in buckets:
            should_remove, why = self._should_remove_bucket(bucket)
            if not should_remove:
                print(f'Skipping bucket removal {bucket.name!r} because {why}')
                remaining_buckets.append(bucket)
                continue

            print('Trying to remove bucket:', bucket.name, 'because', why)
            try:
                self.clean_bucket(bucket)
            except BucketIdNotFound:
                print(f'It seems that bucket {bucket.name} has already been removed')
        print('Total bucket count after cleanup:', len(remaining_buckets))
        for bucket in remaining_buckets:
            print(bucket)

    @backoff.on_exception(
        backoff.expo,
        TooManyRequests,
        max_tries=8,
    )
    def clean_bucket(self, bucket: Bucket | str):
        if isinstance(bucket, str):
            bucket = self.api.get_bucket_by_name(bucket)

        # try optimistic bucket removal first, since it is completely free (as opposed to `ls` call)
        try:
            return self.api.delete_bucket(bucket)
        except (BucketIdNotFound, v3BucketIdNotFound):
            return  # bucket was already removed
        except BadRequest as exc:
            assert exc.code == 'cannot_delete_non_empty_bucket'

        files_leftover = False
        file_versions = bucket.ls(latest_only=False, recursive=True)

        for file_version_info, _ in file_versions:
            if file_version_info.file_retention:
                if file_version_info.file_retention.mode == RetentionMode.GOVERNANCE:
                    print('Removing retention from file version:', file_version_info.id_)
                    self.api.update_file_retention(
                        file_version_info.id_, file_version_info.file_name,
                        NO_RETENTION_FILE_SETTING, True
                    )
                elif file_version_info.file_retention.mode == RetentionMode.COMPLIANCE:
                    if file_version_info.file_retention.retain_until > current_time_millis():  # yapf: disable
                        print(
                            f'File version: {file_version_info.id_} cannot be removed due to compliance mode retention'
                        )
                        files_leftover = True
                        continue
                elif file_version_info.file_retention.mode == RetentionMode.NONE:
                    pass
                else:
                    raise ValueError(
                        f'Unknown retention mode: {file_version_info.file_retention.mode}'
                    )
            if file_version_info.legal_hold.is_on():
                print('Removing legal hold from file version:', file_version_info.id_)
                self.api.update_file_legal_hold(
                    file_version_info.id_, file_version_info.file_name, LegalHold.OFF
                )
            print('Removing file version:', file_version_info.id_)
            try:
                self.api.delete_file_version(file_version_info.id_, file_version_info.file_name)
            except FileNotPresent:
                print(
                    f'It seems that file version {file_version_info.id_} has already been removed'
                )

        if files_leftover:
            print('Unable to remove bucket because some retained files remain')
        else:
            print('Removing bucket:', bucket.name)
            try:
                self.api.delete_bucket(bucket)
            except BucketIdNotFound:
                print(f'It seems that bucket {bucket.name} has already been removed')
        print()

    def count_and_print_buckets(self) -> int:
        buckets = self.api.list_buckets()
        count = len(buckets)
        print(f'Total bucket count at {datetime.now()}: {count}')
        for i, bucket in enumerate(buckets, start=1):
            print(f'- {i}\t{bucket.name} [{bucket.id_}]')
        return count


def print_text_indented(text):
    """
    Prints text that may include weird characters, indented four spaces.
    """
    for line in text.split(linesep):
        Command._print_helper(sys.stdout, sys.stdout.encoding, '   ', repr(line)[1:-1])


def print_output(status, stdout, stderr):
    print('  status:', status)
    if stdout != '':
        print('  stdout:')
        print_text_indented(stdout)
    if stderr != '':
        print('  stderr:')
        print_text_indented(stderr)
    print()


def serialize_enc_settings(value):
    if not isinstance(value, EncryptionSetting):
        raise TypeError
    return value.as_dict()


def print_json_indented(value):
    """
    Converts the value to JSON, then prints it.
    """
    print_text_indented(
        json.dumps(
            value, indent=4, sort_keys=True, ensure_ascii=True, default=serialize_enc_settings
        )
    )


def remove_warnings(text):
    return linesep.join(line for line in text.split(linesep) if 'DeprecationWarning' not in line)


class StringReader:
    def __init__(self):
        self.string = None

    def get_string(self):
        return self.string

    def read_from(self, f):
        try:
            self.string = f.read()
        except Exception as e:
            print(e)
            self.string = str(e)


def should_equal(expected, actual):
    print('  expected:')
    print_json_indented(expected)
    print('  actual:')
    print_json_indented(actual)
    assert expected == actual
    print()


class CommandLine:

    EXPECTED_STDERR_PATTERNS = [
        re.compile(r'^Using https?://[\w.]+$'),  # account auth
        re.compile(r'.*B/s]$', re.DOTALL),  # progress bar
        re.compile(r'^\r?$'),  # empty line
        re.compile(
            r'Encrypting file\(s\) with SSE-C without providing key id. '
            r'Set B2_DESTINATION_SSE_C_KEY_ID to allow key identification'
        ),
        re.compile(
            r'WARNING: Unable to print unicode.  Encoding for stdout is: '
            r'\'[a-zA-Z0-9]+\''
        ),  # windows-bundle tests on CI use cp1252
        re.compile(r'Trying to print: .*'),
    ]

    def __init__(
        self,
        command,
        account_id,
        application_key,
        realm,
        bucket_name_prefix,
        env_file_cmd_placeholder,
        api_wrapper: Api,
        b2_uri_args,
    ):
        self.command = command
        self.account_id = account_id
        self.application_key = application_key
        self.realm = realm
        self.bucket_name_prefix = bucket_name_prefix
        self.env_file_cmd_placeholder = env_file_cmd_placeholder
        self.api_wrapper = api_wrapper
        self.b2_uri_args = b2_uri_args

    def generate_bucket_name(self):
        return self.api_wrapper.new_bucket_name()

    def get_bucket_info_args(self) -> tuple[str, str]:
        return '--bucket-info', json.dumps(self.api_wrapper.new_bucket_info(), ensure_ascii=True)

    def run_command(self, args, additional_env: dict | None = None):
        """
        Runs the command with the given arguments, returns a tuple in form of
        (succeeded, stdout)
        """
        status, stdout, stderr = self.execute(args, additional_env)
        return status == 0 and stderr == '', stdout

    def should_succeed(
        self,
        args: list[str] | None,
        expected_pattern: str | None = None,
        additional_env: dict | None = None,
        expected_stderr_pattern: str | re.Pattern = None,
    ) -> str:
        """
        Runs the command-line with the given arguments.  Raises an exception
        if there was an error; otherwise, returns the stdout of the command
        as string.
        """
        status, stdout, stderr = self.execute(args, additional_env)
        assert status == 0, f'FAILED with status {status}, stderr={stderr}'

        if expected_stderr_pattern:
            assert expected_stderr_pattern.search(stderr), \
                f'stderr did not match pattern="{expected_stderr_pattern}", stderr="{stderr}"'
        elif stderr != '':
            for line in (s.strip() for s in stderr.split(os.linesep)):
                assert any(p.match(line) for p in self.EXPECTED_STDERR_PATTERNS), \
                    f'Unexpected stderr line: {repr(line)}'

        if expected_pattern is not None:
            assert re.search(expected_pattern, stdout), \
            f'did not match pattern="{expected_pattern}", stdout="{stdout}"'

        return stdout.replace(os.linesep, '\n')

    @classmethod
    def prepare_env(self, additional_env: dict | None = None):
        environ['PYTHONIOENCODING'] = 'utf-8'
        env = environ.copy()
        env.update(additional_env or {})
        return env

    def parse_command(self, env):
        """
        Split `self.command` into a list of strings. If necessary, dump the env vars to a tmp file and substitute
        one the command's argument with that file's path.
        """
        command = self.command.split(' ')
        if self.env_file_cmd_placeholder:
            if any('\n' in var_value for var_value in env.values()):
                raise ValueError(
                    'Env vars containing new line characters will break env file format'
                )
            env_file_path = mktemp()
            pathlib.Path(env_file_path).write_text('\n'.join(f'{k}={v}' for k, v in env.items()))
            command = [
                (c if c != self.env_file_cmd_placeholder else env_file_path) for c in command
            ]
        return command

    def execute(
        self,
        args: list[str | Path | int] | None = None,
        additional_env: dict | None = None,
    ):
        """
        :param cmd: a command to run
        :param args: command's arguments
        :param additional_env: environment variables to pass to the command, overwriting parent process ones
        :return: (status, stdout, stderr)
        """
        # We'll run the b2 command-line by running the b2 module from
        # the current directory or provided as parameter
        env = self.prepare_env(additional_env)
        command = self.parse_command(env)

        args: list[str] = [str(arg) for arg in args] if args else []
        command.extend(args)

        print('Running:', ' '.join(command))

        stdout = StringReader()
        stderr = StringReader()

        p = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=platform.system() != 'Windows',
            env=env,
        )
        p.stdin.close()
        reader1 = threading.Thread(target=stdout.read_from, args=[p.stdout])
        reader1.start()
        reader2 = threading.Thread(target=stderr.read_from, args=[p.stderr])
        reader2.start()
        p.wait()
        reader1.join()
        reader2.join()

        stdout_decoded = remove_warnings(stdout.get_string().decode('utf-8', errors='replace'))
        stderr_decoded = remove_warnings(stderr.get_string().decode('utf-8', errors='replace'))

        print_output(p.returncode, stdout_decoded, stderr_decoded)
        return p.returncode, stdout_decoded, stderr_decoded

    def should_succeed_json(self, args, additional_env: dict | None = None, **kwargs):
        """
        Runs the command-line with the given arguments.  Raises an exception
        if there was an error; otherwise, treats the stdout as JSON and returns
        the data in it.
        """
        result = self.should_succeed(args, additional_env=additional_env, **kwargs)
        try:
            loaded_result = json.loads(result)
        except json.JSONDecodeError:
            raise ValueError(f'{result} is not a valid json')
        return loaded_result

    def should_fail(self, args, expected_pattern, additional_env: dict | None = None):
        """
        Runs the command-line with the given args, expecting the given pattern
        to appear in stderr.
        """
        status, stdout, stderr = self.execute(args, additional_env)
        assert status != 0, 'ERROR: should have failed'

        assert re.search(expected_pattern, stdout + stderr), \
            f'did not match pattern="{expected_pattern}", stdout="{stdout}", stderr="{stderr}"'

    def reauthorize(self, check_key_capabilities=False):
        """Clear and authorize again to the account."""
        self.should_succeed(['clear-account'])
        self.should_succeed(
            [
                'authorize-account', '--environment', self.realm, self.account_id,
                self.application_key
            ]
        )
        if check_key_capabilities:
            auth_dict = self.should_succeed_json(['get-account-info'])
            private_preview_caps = {
                'readBucketNotifications',
                'writeBucketNotifications',
            }
            missing_capabilities = set(ALL_CAPABILITIES) - {
                'readBuckets', 'listAllBucketNames'
            } - private_preview_caps - set(auth_dict['allowed']['capabilities'])
            assert not missing_capabilities, f'it appears that the raw_api integration test is being run with a non-full key. Missing capabilities: {missing_capabilities}'

    def list_file_versions(self, bucket_name):
        return self.should_succeed_json(
            ['ls', '--json', '--recursive', '--versions', *self.b2_uri_args(bucket_name)]
        )

    def cleanup_buckets(self, buckets: dict[str, dict | None]) -> None:
        for bucket_name, bucket_dict in buckets.items():
            self.cleanup_bucket(bucket_name, bucket_dict)

    def cleanup_bucket(self, bucket_name: str, bucket_dict: dict | None = None) -> None:
        """
        Cleanup bucket

        Since bucket was being handled by the tool, it is safe to assume it is cached in its cache and we don't
        need to call C class API list_buckets endpoint to get it.
        """
        if not bucket_dict:
            try:
                bucket_dict = self.should_succeed_json(['get-bucket', bucket_name])
            except (ValueError, AssertionError):  # bucket doesn't exist
                return

        bucket = self.api_wrapper.api.BUCKET_CLASS(
            api=self.api_wrapper.api,
            id_=bucket_dict['bucketId'],
            name=bucket_name,
        )
        self.api_wrapper.clean_bucket(bucket)


class TempDir:
    def __init__(self):
        warnings.warn(
            'TempDir is deprecated; use pytest tmp_path fixture instead',
            DeprecationWarning,
            stacklevel=2,
        )
        self.dirpath = None

    def __enter__(self):
        self.dirpath = mkdtemp()
        return Path(self.dirpath)

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(fix_windows_path_limit(self.dirpath))


def read_file(path: str | Path):
    with open(path, 'rb') as f:
        return f.read()


def write_file(path: str | Path, contents: bytes):
    with open(path, 'wb') as f:
        f.write(contents)


def file_mod_time_millis(path: str | Path) -> int:
    return int(os.path.getmtime(path) * 1000)


def set_file_mod_time_millis(path: str | Path, time):
    os.utime(path, (os.path.getatime(path), time / 1000))


def random_hex(length):
    return ''.join(RNG.choice('0123456789abcdef') for _ in range(length))



######################################################################
#
# File: test/integration/cleanup_buckets.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################


def test_cleanup_buckets(b2_api):
    # this is not a test, but it is intended to be called
    # via pytest because it reuses fixtures which have everything
    # set up
    pass  # b2_api calls b2_api.clean_buckets() in its finalizer



######################################################################
#
# File: test/unit/console_tool/test_ls.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
######################################################################
#
# File: test/unit/console_tool/test_download_file.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import pytest


def test_ls__without_bucket_name(b2_cli, bucket_info):
    expected_output = "bucket_0  allPublic   my-bucket\n"

    b2_cli.run(["ls"], expected_stdout=expected_output)
    b2_cli.run(["ls", "b2://"], expected_stdout=expected_output)


def test_ls__without_bucket_name__json(b2_cli, bucket_info):
    expected_output = [
        {
            "accountId": "account-0",
            "bucketId": "bucket_0",
            "bucketInfo": {},
            "bucketName": "my-bucket",
            "bucketType": "allPublic",
            "corsRules": [],
            "defaultRetention": {
                "mode": None
            },
            "defaultServerSideEncryption": {
                "mode": "none"
            },
            "isFileLockEnabled": False,
            "lifecycleRules": [],
            "options": [],
            "replication": {
                "asReplicationDestination": None,
                "asReplicationSource": None,
            },
            "revision": 1,
        }
    ]

    b2_cli.run(["ls", "--json"], expected_json_in_stdout=expected_output)
    b2_cli.run(["ls", "--json", "b2://"], expected_json_in_stdout=expected_output)


@pytest.mark.parametrize("flag", ["--long", "--recursive", "--replication"])
def test_ls__without_bucket_name__option_not_supported(b2_cli, bucket_info, flag):
    b2_cli.run(
        ["ls", flag],
        expected_stderr=f"ERROR: Cannot use {flag} option without specifying a bucket name\n",
        expected_status=1,
    )



######################################################################
#
# File: test/unit/console_tool/conftest.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import os
import sys

import pytest

import b2._internal.console_tool


@pytest.fixture
def cwd_path(tmp_path):
    """Set up a test directory and return its path."""
    prev_cwd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(prev_cwd)


@pytest.fixture
def b2_cli_log_fix(caplog):
    caplog.set_level(0)  # prevent pytest from blocking logs
    b2._internal.console_tool.logger.setLevel(0)  # reset logger level to default


@pytest.fixture
def mock_stdin(monkeypatch):
    out_, in_ = os.pipe()
    monkeypatch.setattr(sys, 'stdin', os.fdopen(out_))
    in_f = open(in_, 'w')
    yield in_f
    in_f.close()



######################################################################
#
# File: test/integration/test_autocomplete.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import sys
from test.helpers import skip_on_windows

import pexpect
import pytest

TIMEOUT = 120  # CI can be slow at times when parallelization is extreme

BASHRC_CONTENT = """\
# ~/.bashrc dummy file

echo "Just testing if we don't replace existing script" > /dev/null
# >>> just a test section >>>
# regardless what is in there already
# <<< just a test section <<<
"""


@pytest.fixture(scope="session")
def bashrc(homedir):
    bashrc_path = (homedir / '.bashrc')
    bashrc_path.write_text(BASHRC_CONTENT)
    yield bashrc_path


@pytest.fixture(scope="module")
def cli_command(request) -> str:
    return request.config.getoption('--sut')


@pytest.fixture(scope="module")
def autocomplete_installed(env, homedir, bashrc, cli_version, cli_command, is_running_on_docker):
    if is_running_on_docker:
        pytest.skip('Not supported on Docker')

    shell = pexpect.spawn(
        f'bash -i -c "{cli_command} install-autocomplete"', env=env, logfile=sys.stderr.buffer
    )
    try:
        shell.expect_exact('Autocomplete successfully installed for bash', timeout=TIMEOUT)
    finally:
        shell.close()
    shell.wait()
    assert (homedir / '.bash_completion.d' / cli_version).is_file()
    assert bashrc.read_text().startswith(BASHRC_CONTENT)


@pytest.fixture
def shell(env):
    shell = pexpect.spawn('bash -i', env=env, maxread=1000)
    shell.setwinsize(100, 100)  # required to see all suggestions in tests
    yield shell
    shell.close()


@skip_on_windows
def test_autocomplete_b2_commands(autocomplete_installed, is_running_on_docker, shell, cli_version):
    if is_running_on_docker:
        pytest.skip('Not supported on Docker')
    shell.send(f'{cli_version} \t\t')
    shell.expect_exact(["authorize-account", "download-file", "get-bucket"], timeout=TIMEOUT)


@skip_on_windows
def test_autocomplete_b2_only_matching_commands(
    autocomplete_installed, is_running_on_docker, shell, cli_version
):
    if is_running_on_docker:
        pytest.skip('Not supported on Docker')
    shell.send(f'{cli_version} delete-\t\t')

    shell.expect_exact("file", timeout=TIMEOUT)  # common part of remaining cmds is autocompleted
    with pytest.raises(pexpect.exceptions.TIMEOUT):  # no other commands are suggested
        shell.expect_exact("get-bucket", timeout=0.5)


@skip_on_windows
def test_autocomplete_b2__download_file__b2uri(
    autocomplete_installed,
    shell,
    b2_tool,
    bucket_name,
    file_name,
    is_running_on_docker,
    cli_version,
):
    """Test that autocomplete suggests bucket names and file names."""
    if is_running_on_docker:
        pytest.skip('Not supported on Docker')
    shell.send(f'{cli_version} download_file \t\t')
    shell.expect_exact("b2://", timeout=TIMEOUT)
    shell.send('b2://\t\t')
    shell.expect_exact(bucket_name, timeout=TIMEOUT)
    shell.send(f'{bucket_name}/\t\t')
    shell.expect_exact(file_name, timeout=TIMEOUT)



######################################################################
#
# File: test/unit/console_tool/test_download_file.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import os
import pathlib
from test.helpers import skip_on_windows

import pytest

EXPECTED_STDOUT_DOWNLOAD = '''
File name:           file1.txt
File id:             9999
Output file path:    {output_path}
File size:           11
Content type:        b2/x-auto
Content sha1:        2aae6c35c94fcfb415dbe95f408b9ce91ee846ed
Encryption:          none
Retention:           none
Legal hold:          <unset>
INFO src_last_modified_millis: 1500111222000
Checksum matches
Download finished
'''


@pytest.mark.parametrize(
    'flag,expected_stdout', [
        ('--no-progress', EXPECTED_STDOUT_DOWNLOAD),
        ('-q', ''),
        ('--quiet', ''),
    ]
)
def test_download_file_by_uri__flag_support(b2_cli, uploaded_file, tmp_path, flag, expected_stdout):
    output_path = tmp_path / 'output.txt'

    b2_cli.run(
        ['download-file', flag, 'b2id://9999',
         str(output_path)],
        expected_stdout=expected_stdout.format(output_path=pathlib.Path(output_path).resolve())
    )
    assert output_path.read_text() == uploaded_file['content']


@pytest.mark.parametrize('b2_uri', [
    'b2://my-bucket/file1.txt',
    'b2id://9999',
])
def test_download_file_by_uri__b2_uri_support(b2_cli, uploaded_file, tmp_path, b2_uri):
    output_path = tmp_path / 'output.txt'

    b2_cli.run(
        ['download-file', b2_uri, str(output_path)],
        expected_stdout=EXPECTED_STDOUT_DOWNLOAD.format(
            output_path=pathlib.Path(output_path).resolve()
        )
    )
    assert output_path.read_text() == uploaded_file['content']


@pytest.mark.parametrize(
    'flag,expected_stdout', [
        ('--no-progress', EXPECTED_STDOUT_DOWNLOAD),
        ('-q', ''),
        ('--quiet', ''),
    ]
)
def test_download_file_by_name(b2_cli, local_file, uploaded_file, tmp_path, flag, expected_stdout):
    output_path = tmp_path / 'output.txt'

    b2_cli.run(
        [
            'download-file-by-name', uploaded_file['bucket'], uploaded_file['fileName'],
            str(output_path)
        ],
        expected_stdout=EXPECTED_STDOUT_DOWNLOAD.format(
            output_path=pathlib.Path(output_path).resolve()
        ),
        expected_stderr=
        'WARNING: download-file-by-name command is deprecated. Use download-file instead.\n',
    )
    assert output_path.read_text() == uploaded_file['content']


@pytest.mark.parametrize(
    'flag,expected_stdout', [
        ('--no-progress', EXPECTED_STDOUT_DOWNLOAD),
        ('-q', ''),
        ('--quiet', ''),
    ]
)
def test_download_file_by_id(b2_cli, uploaded_file, tmp_path, flag, expected_stdout):
    output_path = tmp_path / 'output.txt'

    b2_cli.run(
        ['download-file-by-id', flag, '9999', str(output_path)],
        expected_stdout=expected_stdout.format(output_path=pathlib.Path(output_path).resolve()),
        expected_stderr=
        'WARNING: download-file-by-id command is deprecated. Use download-file instead.\n',
    )
    assert output_path.read_text() == uploaded_file['content']


@skip_on_windows(reason='os.mkfifo is not supported on Windows')
def test_download_file_by_name__named_pipe(
    b2_cli, local_file, uploaded_file, tmp_path, bg_executor
):
    output_path = tmp_path / 'output.txt'
    os.mkfifo(output_path)

    output_string = None

    def reader():
        nonlocal output_string
        output_string = output_path.read_text()

    reader_future = bg_executor.submit(reader)

    b2_cli.run(
        [
            'download-file-by-name', '--no-progress', uploaded_file['bucket'],
            uploaded_file['fileName'],
            str(output_path)
        ],
        expected_stdout=EXPECTED_STDOUT_DOWNLOAD.format(
            output_path=pathlib.Path(output_path).resolve()
        ),
        expected_stderr=
        'WARNING: download-file-by-name command is deprecated. Use download-file instead.\n',
    )
    reader_future.result(timeout=1)
    assert output_string == uploaded_file['content']


@pytest.fixture
def uploaded_stdout_txt(b2_cli, bucket, local_file, tmp_path):
    local_file.write_text('non-mocked /dev/stdout test ignore me')
    b2_cli.run(['upload-file', bucket, str(local_file), 'stdout.txt'])
    return {
        'bucket': bucket,
        'fileName': 'stdout.txt',
        'content': local_file.read_text(),
    }


def test_download_file_by_name__to_stdout_by_alias(
    b2_cli, bucket, uploaded_stdout_txt, tmp_path, capfd
):
    """Test download_file_by_name stdout alias support"""
    b2_cli.run(
        ['download-file-by-name', '--no-progress', bucket, uploaded_stdout_txt['fileName'], '-'],
        expected_stderr=
        'WARNING: download-file-by-name command is deprecated. Use download-file instead.\n',
    )
    assert capfd.readouterr().out == uploaded_stdout_txt['content']
    assert not pathlib.Path('-').exists()


def test_cat__b2_uri(b2_cli, bucket, uploaded_stdout_txt, tmp_path, capfd):
    """Test download_file_by_name stdout alias support"""
    b2_cli.run(['cat', '--no-progress', f"b2://{bucket}/{uploaded_stdout_txt['fileName']}"],)
    assert capfd.readouterr().out == uploaded_stdout_txt['content']


def test_cat__b2_uri__invalid(b2_cli, capfd):
    b2_cli.run(
        ['cat', "nothing/meaningful"],
        expected_stderr=None,
        expected_status=2,
    )
    assert "argument B2_URI: Unsupported URI scheme: ''" in capfd.readouterr().err


def test_cat__b2_uri__not_a_file(b2_cli, bucket, capfd):
    b2_cli.run(
        ['cat', "b2://bucket/dir/subdir/"],
        expected_stderr=None,
        expected_status=2,
    )
    assert "argument B2_URI: B2 URI pointing to a file-like object is required" in capfd.readouterr(
    ).err


def test_cat__b2id_uri(b2_cli, bucket, uploaded_stdout_txt, tmp_path, capfd):
    """Test download_file_by_name stdout alias support"""
    b2_cli.run(['cat', '--no-progress', "b2id://9999"],)
    assert capfd.readouterr().out == uploaded_stdout_txt['content']


def test__download_file__threads(b2_cli, local_file, uploaded_file, tmp_path):
    num_threads = 13
    output_path = tmp_path / 'output.txt'

    b2_cli.run(
        [
            'download-file', '--no-progress', '--threads',
            str(num_threads), 'b2://my-bucket/file1.txt',
            str(output_path)
        ]
    )

    assert output_path.read_text() == uploaded_file['content']
    assert b2_cli.console_tool.api.services.download_manager.get_thread_pool_size() == num_threads



######################################################################
#
# File: b2/_internal/_cli/shell.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import os
import os.path
import shutil
from typing import Optional


def detect_shell() -> Optional[str]:
    """Detect the shell we are running in."""
    shell_var = os.environ.get('SHELL')
    if shell_var:
        return os.path.basename(shell_var)
    return None


def resolve_short_call_name(binary_path: str) -> str:
    """
    Resolve the short name of the binary.

    If binary is in PATH, return only basename, otherwise return a full path.
    This method is to be used with sys.argv[0] to resolve handy name for the user instead of full path.
    """
    if shutil.which(binary_path) == binary_path:
        return os.path.basename(binary_path)
    return binary_path



######################################################################
#
# File: test/conftest.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import sys

import pytest

from b2._internal._utils.python_compat import removeprefix


@pytest.hookimpl
def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "apiver(from_ver, to_ver): run tests only on certain apiver versions",
    )


@pytest.fixture(scope='session')
def apiver(request):
    """Get apiver as a v-prefixed string, e.g. "v2"."""
    return removeprefix(request.config.getoption('--cli', '').lstrip('_'), "b2") or None


@pytest.fixture(scope='session')
def apiver_int(apiver) -> int:
    return int(apiver[1:]) if apiver else -1


@pytest.fixture(autouse=True)
def run_on_apiver_handler(request, apiver_int):
    """
    Auto-fixture that allows skipping tests based on the CLI apiver versions.

    Usage:
        @pytest.mark.apiver(1, 3)
        def test_foo():
            # Test is run only for versions 1 and 3
            ...

        @pytest.mark.apiver(from_ver=2, to_ver=5)
        def test_bar():
            # Test is run only for versions 2, 3, 4 and 5
            ...

    Note that it requires the `cli_int_version` fixture to be defined.
    Both unit tests and integration tests handle it a little bit different, thus
    two different fixtures are provided.
    """
    node = request.node.get_closest_marker('apiver')
    if not node:
        return

    if not node.args and not node.kwargs:
        return

    assert apiver_int >= 0, 'apiver_int fixture is not defined'

    if node.args:
        if apiver_int in node.args:
            # Run the test.
            return

    if node.kwargs:
        from_ver = node.kwargs.get('from_ver', 0)
        to_ver = node.kwargs.get('to_ver', sys.maxsize)

        if from_ver <= apiver_int <= to_ver:
            # Run the test.
            return

    pytest.skip('Not supported on this apiver version')



######################################################################
#
# File: pyinstaller-hooks/hook-prettytable.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

from PyInstaller.utils.hooks import collect_all

# prettytable is excluded because `prettytable` module in provided by `PTable` package;
# pyinstaller fails to resolve this, thus we do it manually here
excludedimports = ['prettytable']
datas, binaries, hiddenimports = collect_all('prettytable')



######################################################################
#
# File: noxfile.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import datetime
import hashlib
import os
import pathlib
import platform
import re
import string
import subprocess

import nox

# Required for PDM to use nox's virtualenvs
os.environ["PDM_IGNORE_SAVED_PYTHON"] = "1"
os.environ["PDM_NO_LOCK"] = "1"

UPSTREAM_REPO_URL = 'git@github.com:Backblaze/B2_Command_Line_Tool.git'

CI = os.environ.get('CI') is not None
CD = CI and (os.environ.get('CD') is not None)
INSTALL_SDK_FROM = os.environ.get('INSTALL_SDK_FROM')
NO_STATICX = os.environ.get('NO_STATICX') is not None
NOX_PYTHONS = os.environ.get('NOX_PYTHONS')

PYTHON_VERSIONS = [
    'pypy3.9',
    'pypy3.10',
    '3.7',
    '3.8',
    '3.9',
    '3.10',
    '3.11',
    '3.12',
] if NOX_PYTHONS is None else NOX_PYTHONS.split(',')


def _detect_python_nox_id() -> str:
    major, minor, *_ = platform.python_version_tuple()
    python_nox_id = f"{major}.{minor}"
    if platform.python_implementation() == 'PyPy':
        python_nox_id = f"pypy{python_nox_id}"
    return python_nox_id


if CI and not NOX_PYTHONS:
    # this is done to allow it to work even if `nox -p` was passed to nox
    PYTHON_VERSIONS = [_detect_python_nox_id()]
    print(f"CI job mode; using provided interpreter only; PYTHON_VERSIONS={PYTHON_VERSIONS!r}")

PYTHON_DEFAULT_VERSION = PYTHON_VERSIONS[-1]

PY_PATHS = ['b2', 'test', 'noxfile.py']

DOCKER_TEMPLATE = pathlib.Path('./Dockerfile.template')

SYSTEM = platform.system().lower()

WINDOWS_TIMESTAMP_SERVER = 'http://timestamp.digicert.com'
WINDOWS_SIGNTOOL_PATH = 'C:/Program Files (x86)/Windows Kits/10/bin/10.0.17763.0/x86/signtool.exe'

nox.options.reuse_existing_virtualenvs = not CI
nox.options.sessions = [
    'lint',
    'test',
]

PYTEST_GLOBAL_ARGS = []
if CI:
    PYTEST_GLOBAL_ARGS.append("-vv")


def pdm_install(
    session: nox.Session, *groups: str, dev: bool = True, editable: bool = False
) -> None:
    args = []
    if not dev:
        args.append('--prod')
    if not editable:
        args.append('--no-editable')
    for group in groups:
        args.extend(['--group', group])
    session.run('pdm', 'install', *args, external=True)


def github_output(name, value, *, secret=False):
    gh_output_path = os.environ.get('GITHUB_OUTPUT')
    if secret:
        print(f"::add-mask::{value}")
    if gh_output_path:
        with open(gh_output_path, "a") as file:
            file.write(f"{name}={value}\n")
    else:
        print(f"github_output {name}={'******' if secret else value}")


def get_version_key(path: pathlib.Path) -> int:
    version_name = path.name
    # There is no version 0, thus we can provide it to the element starting with an underscore.
    if version_name.startswith('_'):
        return 0

    version_match = re.match(r'[_]*b2v(\d+)', version_name)
    assert version_match, f'Version {version_name} does not match pattern B2Cli version pattern.'
    version_number = int(version_match.group(1))
    return version_number


def get_versions() -> list[str]:
    """
    "Almost" a copy of b2/_internal/version_listing.py:get_versions(), because importing
    the file directly seems impossible from the noxfile.
    """
    # This sorting ensures that:
    # - the first element is the latest unstable version (starts with an underscore)
    # - the last element is the latest stable version (highest version number)
    return [
        path.name for path in sorted(
            (pathlib.Path(__file__).parent / 'b2' / '_internal').glob('*b2v*'),
            key=get_version_key,
        )
    ]


@nox.session(name='format', python=PYTHON_DEFAULT_VERSION)
def format_(session):
    """Lint the code and apply fixes in-place whenever possible."""
    pdm_install(session, 'format')
    # TODO: incremental mode for yapf
    session.run('yapf', '--in-place', '--parallel', '--recursive', *PY_PATHS)
    session.run('ruff', 'check', '--fix', *PY_PATHS)
    # session.run(
    #     'docformatter',
    #     '--in-place',
    #     '--recursive',
    #     '--wrap-summaries=100',
    #     '--wrap-descriptions=100',
    #     *PY_PATHS,
    # )


@nox.session(python=PYTHON_DEFAULT_VERSION)
def lint(session):
    """Run linters in readonly mode."""
    pdm_install(session, 'lint', 'doc', 'full', 'license')
    session.run('yapf', '--diff', '--parallel', '--recursive', *PY_PATHS)
    session.run('ruff', 'check', *PY_PATHS)
    # session.run(
    #     'docformatter',
    #     '--check',
    #     '--recursive',
    #     '--wrap-summaries=100',
    #     '--wrap-descriptions=100',
    #     *PY_PATHS,
    # )

    session.run('pytest', 'test/static', *PYTEST_GLOBAL_ARGS)
    session.run('liccheck', '-s', 'pyproject.toml')
    session.run('pdm', 'lock', '--check', external=True)


@nox.session(python=PYTHON_VERSIONS)
def unit(session):
    """Run unit tests."""
    pdm_install(session, 'test')

    command = [
        'pytest',
        '-n',
        'auto',
        '--cov=b2',
        '--cov-branch',
        '--cov-report=xml',
        '--doctest-modules',
        *PYTEST_GLOBAL_ARGS,
        *session.posargs,
        'test/unit',
    ]

    versions = get_versions()
    session.run(*command, '--cli', versions[0])
    command.append('--cov-append')
    if not session.posargs:
        session.notify('cover')

    for cli_version in versions[1:]:
        session.run(*command, '--cli', cli_version)


def run_integration_test(session, pytest_posargs):
    """Run integration tests."""
    pdm_install(session, 'license', 'test')

    command = [
        'pytest',
        'test/integration',
        '-s',
        '-n',
        '2' if CI else 'auto',
        '--log-level',
        'INFO',
        '-W',
        'ignore::DeprecationWarning:rst2ansi.visitor:',
        *PYTEST_GLOBAL_ARGS,
        *pytest_posargs,
    ]

    # sut can be provided explicitly (like in docker) or like `"--sut=path/b2"`.
    provided_sut = any('--sut' in elem for elem in pytest_posargs)

    # If `sut` was provided, we just run this one.
    # If not, we're running the test on all known versions.
    if provided_sut:
        session.run(*command)
    else:
        versions = get_versions()
        for cli_version in versions:
            # If we're in a virtualenv, we want to extract the path to the executable
            # that's installed in the virtualenv.  This may not be elegant but shutil
            # gives us a cross-platform solution out of the box.
            exe_path = session.run(
                'python', '-c', f'import shutil; print(shutil.which("{cli_version}"))', silent=True
            ).strip()
            session.run(*command, '--sut', exe_path)


@nox.session(python=PYTHON_VERSIONS)
def integration(session):
    """Run integration tests."""
    run_integration_test(session, session.posargs)


@nox.session(python=PYTHON_VERSIONS)
def test(session):
    """Run all tests."""
    if session.python:
        session.notify(f'unit-{session.python}')
        session.notify(f'integration-{session.python}')
    else:
        session.notify('unit')
        session.notify('integration')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def cleanup_buckets(session):
    """Remove buckets from previous test runs."""
    pdm_install(session, 'test')
    session.run(
        'pytest', '-s', '-x', *PYTEST_GLOBAL_ARGS, *session.posargs,
        'test/integration/cleanup_buckets.py'
    )


@nox.session
def cover(session):
    """Perform coverage analysis."""
    pdm_install(session, 'test')
    session.run('coverage', 'report', '--fail-under=75', '--show-missing', '--skip-covered')
    session.run('coverage', 'erase')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def build(session):
    """Build the distribution."""
    session.run('nox', '-s', 'dump_license', '-fb', 'venv', external=True)
    session.run('pdm', 'build', external=True)

    # Path have to be specified with unix style slashes even for windows,
    # otherwise glob won't find files on windows in action-gh-release.
    github_output('asset_path', 'dist/*')

    version = os.environ['GITHUB_REF'].replace('refs/tags/v', '')
    github_output('version', version)


@nox.session(python=PYTHON_DEFAULT_VERSION)
def dump_license(session: nox.Session):
    pdm_install(session, 'license', editable=True)
    session.run('b2', 'license', '--dump', '--with-packages')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def bundle(session: nox.Session):
    """Bundle the distribution."""

    # We're running dump_license in another session because:
    # 1. `b2 license --dump` dumps the licence where the module is installed.
    # 2. We don't want to install b2 as editable module in the current session
    #    because that would make `b2 versions` show the versions as editable.
    session.run('nox', '-s', 'dump_license', '-fb', 'venv', external=True)
    pdm_install(session, 'bundle', 'full')

    template_spec = string.Template(pathlib.Path('b2.spec.template').read_text())
    versions = get_versions()

    # It is assumed that the last element will be the "latest stable".
    for binary_name, version in [('b2', versions[-1])] + list(zip(versions, versions)):
        spec = template_spec.safe_substitute({
            'VERSION': version,
            'NAME': binary_name,
        })
        pathlib.Path(f'{binary_name}.spec').write_text(spec)

        session.run('pyinstaller', *session.posargs, f'{binary_name}.spec')

        if SYSTEM == 'linux' and not NO_STATICX:
            session.run(
                'staticx', '--no-compress', '--strip', '--loglevel', 'INFO', f'dist/{binary_name}',
                f'dist/{binary_name}-static'
            )
            session.run(
                'mv',
                '-f',
                f'dist/{binary_name}-static',
                f'dist/{binary_name}',
                external=True,
            )

    # Path have to be specified with unix style slashes even for windows,
    # otherwise glob won't find files on windows in action-gh-release.
    github_output('asset_path', 'dist/*')

    # Note: this should pick the shortest named executable from the directory.
    # But, for yet unknown reason, the `./dist/b2` doesn't play well with `--sut` and the autocomplete.
    # For this reason, we're returning here the "latest, stable version" instead.
    # This current implementation works fine up until version 10, when it will break.
    # By that time, we should have come back to picking the shortest named binary (`b2`) up.
    executable = max(
        str(path) for path in pathlib.Path('dist').glob('*') if not path.name.startswith('_')
    )
    github_output('sut_path', executable)


@nox.session(python=False)
def sign(session):
    """Sign the bundled distribution (macOS and Windows only)."""

    def sign_windows(cert_file, cert_password):
        session.run('certutil', '-f', '-p', cert_password, '-importpfx', cert_file)
        for binary_name in ['b2'] + get_versions():
            session.run(
                WINDOWS_SIGNTOOL_PATH,
                'sign',
                '/f',
                cert_file,
                '/p',
                cert_password,
                '/tr',
                WINDOWS_TIMESTAMP_SERVER,
                '/td',
                'sha256',
                '/fd',
                'sha256',
                f'dist/{binary_name}.exe',
                external=True,
            )
            session.run(
                WINDOWS_SIGNTOOL_PATH,
                'verify',
                '/pa',
                '/all',
                f'dist/{binary_name}.exe',
                external=True,
            )

    if SYSTEM == 'windows':
        try:
            certificate_file, certificate_password = session.posargs
        except ValueError:
            session.error('pass the certificate file and the password as positional arguments')
            return

        sign_windows(certificate_file, certificate_password)
    elif SYSTEM == 'linux':
        session.log('signing is not supported for Linux')
    else:
        session.error(f'unrecognized platform: {SYSTEM}')

    # Append OS name to all the binaries.
    for asset in pathlib.Path('dist').glob('*'):
        name = asset.stem
        ext = asset.suffix
        asset_path = f'dist/{name}-{SYSTEM}{ext}'
        session.run('mv', '-f', asset, asset_path, external=True)

    # Path have to be specified with unix style slashes even for windows,
    # otherwise glob won't find files on windows in action-gh-release.
    github_output('asset_path', 'dist/*')


def _calculate_hashes(
    file_path: pathlib.Path,
    algorithms: list[str],
) -> list[hashlib._Hash]:  # noqa
    read_size = 1024 * 1024
    hash_structures = [hashlib.new(algo) for algo in algorithms]

    with open(file_path, 'rb') as f:
        while True:
            buffer = f.read(read_size)
            if not buffer:
                break

            for hash_struct in hash_structures:
                hash_struct.update(buffer)

    return hash_structures


def _save_hashes(output_file: pathlib.Path, hashes: list[hashlib._Hash]) -> None:  # noqa
    longest_algo_name = max(len(elem.name) for elem in hashes)
    line_format = '{algo:<%s} {hash_value}' % longest_algo_name

    output_lines = []
    for hash_struct in hashes:
        hash_value = hash_struct.hexdigest()
        output_lines.append(line_format.format(algo=hash_struct.name, hash_value=hash_value))

    output_file.write_bytes('\n'.join(output_lines).encode('ascii'))


@nox.session(python=PYTHON_DEFAULT_VERSION)
def make_dist_digest(_session):
    wanted_algos = ['sha256', 'sha512', 'sha3_256', 'sha3_512']
    available_algos = [algo for algo in wanted_algos if algo in hashlib.algorithms_available]

    directory = pathlib.Path('dist')
    glob_match = '*'

    hashes_file_suffix = '_hashes'
    did_find_any_file = False

    for dist_file in directory.glob(glob_match):
        if dist_file.stem.endswith(hashes_file_suffix):
            continue

        hashes_list = _calculate_hashes(dist_file, available_algos)

        output_file = dist_file.with_stem(dist_file.name + hashes_file_suffix).with_suffix('.txt')
        _save_hashes(output_file, hashes_list)

        did_find_any_file = True

    if not did_find_any_file:
        raise RuntimeError(
            f'No file found in {str(directory / glob_match)}, but was expected to find some.'
        )


@nox.session(python=PYTHON_DEFAULT_VERSION)
def doc(session):
    """Build the documentation."""
    pdm_install(session, 'doc')
    session.cd('doc')
    sphinx_args = ['-b', 'html', '-T', '-W', 'source', 'build/html']
    session.run('rm', '-rf', 'build', external=True)

    if not session.interactive:
        session.run('sphinx-build', *sphinx_args)
        # session.notify('doc_cover')  #  disabled due to https://github.com/sphinx-doc/sphinx/issues/11678
    else:
        sphinx_args[-2:-2] = [
            '-E', '--open-browser', '--watch', '../b2', '--ignore', '*.pyc', '--ignore', '*~'
        ]
        session.run('sphinx-autobuild', *sphinx_args)


@nox.session
def doc_cover(session):
    """
    Perform coverage analysis for the documentation.

    At the time of writing B2 CLI does not have object documentation, hence this always returns 0 out 0 objects.
    Which errors out in Sphinx 7.2 (https://github.com/sphinx-doc/sphinx/issues/11678).
    """
    pdm_install(session, 'doc')
    session.cd('doc')
    sphinx_args = ['-b', 'coverage', '-T', '-W', 'source', 'build/coverage']
    session.run('sphinx-build', *sphinx_args)


def _read_readme_name_and_description() -> tuple[str, str]:
    """
    Get name and the description from the readme. First line is assumed to be the project name,
    second contains list of all different checks. Third one and the following contains some description.
    We assume that description can be multiline, and it ends with an empty line.

    An example of the content from README.md can look like this:

    ..note:
        # B2 Command Line Tool
        &nbsp;[![Continuous Integration](https://github.com/Backblaze/B2_Command_Line_Tool/ ... (a very long line)

        (a few empty lines)

        The command-line tool that gives easy access to all of the capabilities of B2 Cloud Storage.

        This program provides command-line access to the B2 service.

    From this we should parse the following:
    "B2 Command Line Tool" as the name and
    "The command-line tool that gives easy access to all of the capabilities of B2 Cloud Storage." as the description.
    """
    with open('README.md') as f:
        non_empty_lines = 0
        full_name = None
        description_parts = []

        for line_with_ends in f.readlines():
            line = line_with_ends.strip()
            if len(line) == 0:
                # If we found an empty line after we got anything for our description – finish.
                if len(description_parts) > 0:
                    break
                continue

            non_empty_lines += 1

            if non_empty_lines == 1:
                # Markdown header starts with some "# ", we strip everything up to first space.
                full_name = line.split(' ', maxsplit=1)[1]

            if non_empty_lines < 3:
                continue

            description_parts.append(line)

    return full_name, ' '.join(description_parts)


@nox.session(python=PYTHON_DEFAULT_VERSION)
def generate_dockerfile(session):
    """Generate Dockerfile from Dockerfile.template"""
    build(session)
    pdm_install(session)

    # This string is like `b2 command line tool, version <sem-ver-string>`
    version = session.run('b2', 'version', '--short', silent=True).strip()

    dist_path = 'dist'

    full_name, description = _read_readme_name_and_description()
    vcs_ref = session.run("git", "rev-parse", "HEAD", external=True, silent=True).strip()
    built_distribution = list(pathlib.Path('.').glob(f'{dist_path}/*'))[0]

    template_mapping = dict(
        python_version=PYTHON_DEFAULT_VERSION,
        vendor='Backblaze',
        name=full_name,
        description=description,
        version=version,
        url='https://www.backblaze.com',
        # TODO: consider fetching it from `git ls-remote --get-url origin`
        vcs_url='https://github.com/Backblaze/B2_Command_Line_Tool',
        vcs_ref=vcs_ref,
        build_date=datetime.datetime.utcnow().isoformat(),
        tar_path=dist_path,
        tar_name=built_distribution.name,
    )

    template_file = DOCKER_TEMPLATE.read_text()
    template = string.Template(template_file)
    dockerfile = template.substitute(template_mapping)
    pathlib.Path('./Dockerfile').write_text(dockerfile)


def run_docker_tests(session, image_tag):
    """Run unittests against a docker image."""
    user_id = session.run('id', '-u', silent=True, external=True).strip()
    group_id = session.run('id', '-g', silent=True, external=True).strip()
    docker_run_cmd = f"docker run -i --user {user_id}:{group_id} -v /tmp:/tmp:rw --env-file ENVFILE"
    run_integration_test(
        session, [
            "--sut",
            f"{docker_run_cmd} {image_tag}",
            "--env-file-cmd-placeholder",
            "ENVFILE",
        ]
    )
    for binary_name in get_versions():
        run_integration_test(
            session, [
                "--sut",
                f"{docker_run_cmd} --entrypoint {binary_name} {image_tag}",
                "--env-file-cmd-placeholder",
                "ENVFILE",
            ]
        )


@nox.session(python=PYTHON_DEFAULT_VERSION)
def docker_test(session):
    """Run unittests against a docker image."""
    if session.posargs:
        image_tag = session.posargs[0]
    else:
        raise ValueError('Provide -- {docker_image_tag}')
    run_docker_tests(session, image_tag)


@nox.session(python=PYTHON_DEFAULT_VERSION)
def build_and_test_docker(session):
    """
    For running locally, CI uses a different set of sessions
    """
    test_image_tag = 'b2:test'
    generate_dockerfile(session)
    session.run('docker', 'build', '-t', test_image_tag, '.', external=True)
    run_docker_tests(session, test_image_tag)


@nox.session(python=PYTHON_DEFAULT_VERSION)
def make_release_commit(session):
    """
    Runs `towncrier build`, commits changes, tags, all that is left to do is pushing
    """
    if session.posargs:
        version = session.posargs[0]
    else:
        session.error('Provide -- {release_version} (X.Y.Z - without leading "v")')

    requirements = session.run('pdm', 'export', '--no-hashes', silent=True)
    # if b2sdk requirement points to git, it won't have a version definition b2sdk==
    assert ('b2sdk==' in requirements) and (
        'git+' not in requirements
    ), 'release version must depend on released b2sdk version'

    if not re.match(r'^\d+\.\d+\.\d+$', version):
        session.error(
            f'Provided version="{version}". Version must be of the form X.Y.Z where '
            f'X, Y and Z are integers'
        )

    local_changes = subprocess.check_output(['git', 'diff', '--stat'])
    if local_changes:
        session.error('Uncommitted changes detected')

    current_branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode()
    if current_branch != 'master':
        session.log('WARNING: releasing from a branch different than master')

    pdm_install(session, 'release')
    session.run('towncrier', 'build', '--yes', '--version', version)

    session.log(
        f'CHANGELOG updated, changes ready to commit and push\n'
        f'    git remote add upstream {UPSTREAM_REPO_URL!r} 2>/dev/null || git remote get-url upstream\n'
        f'    git commit -m "release {version}"\n'
        f'    git tag v{version}\n'
        f'    git push upstream v{version}\n'
        f'    git push upstream {current_branch}'
    )


def load_allowed_change_types(project_toml: pathlib.Path = pathlib.Path('./pyproject.toml')
                             ) -> set[str]:
    """
    Load the list of allowed change types from the pyproject.toml file.
    """
    import tomllib
    configuration = tomllib.loads(project_toml.read_text())
    return set(entry['directory'] for entry in configuration['tool']['towncrier']['type'])


def is_changelog_filename_valid(filename: str, allowed_change_types: set[str]) -> tuple[bool, str]:
    """
    Validates whether the given filename matches our rules.
    Provides information about why it doesn't match them.
    """
    error_reasons = []

    wanted_extension = 'md'
    try:
        description, change_type, extension = filename.rsplit('.', maxsplit=2)
    except ValueError:
        # Not enough values to unpack.
        return False, "Doesn't follow the \"<description>.<change_type>.md\" pattern."

    # Check whether the filename ends with .md.
    if extension != wanted_extension:
        error_reasons.append(f"Doesn't end with {wanted_extension} extension.")

    # Check whether the change type is valid.
    if change_type not in allowed_change_types:
        error_reasons.append(
            f"Change type '{change_type}' doesn't match allowed types: {allowed_change_types}."
        )

    # Check whether the description makes sense.
    try:
        int(description)
    except ValueError:
        if description[0] != '+':
            error_reasons.append("Doesn't start with a number nor a plus sign.")

    return len(error_reasons) == 0, ' / '.join(error_reasons) if error_reasons else ''


def is_changelog_entry_valid(file_content: str) -> tuple[bool, str]:
    """
    We expect the changelog entry to be a valid sentence in the English language.
    This includes, but not limits to, providing a capital letter at the start
    and the full-stop character at the end.

    Note: to do this "properly", tools like `nltk` and `spacy` should be used.
    """
    error_reasons = []

    # Check whether the first character is a capital letter.
    # Not allowing special characters nor numbers at the very start.
    if not file_content[0].isalpha() or not file_content[0].isupper():
        error_reasons.append('The first character is not a capital letter.')

    # Check if the last character is a full-stop character.
    if file_content.strip()[-1] != '.':
        error_reasons.append('The last character is not a full-stop character.')

    return len(error_reasons) == 0, ' / '.join(error_reasons) if error_reasons else ''


@nox.session(python=PYTHON_DEFAULT_VERSION)
def towncrier_check(session):
    """
    Check whether all the entries in the changelog.d follow the expected naming convention
    as well as some basic rules as to their format.
    """
    expected_non_md_files = {'.gitkeep'}
    allowed_change_types = load_allowed_change_types()

    is_error = False

    for filename in pathlib.Path('./changelog.d/').glob('*'):
        # If that's an expected file, it's all right.
        if filename.name in expected_non_md_files:
            continue

        # Check whether the file matches the expected pattern.
        is_valid, error_message = is_changelog_filename_valid(filename.name, allowed_change_types)
        if not is_valid:
            session.log(f"File {filename.name} doesn't match the expected pattern: {error_message}")
            is_error = True
            continue

        # Check whether the file isn't too big.
        if filename.lstat().st_size > 16 * 1024:
            session.log(
                f'File {filename.name} content is too big – it should be smaller than 16kB.'
            )
            is_error = True
            continue

        # Check whether the file can be loaded as UTF-8 file.
        try:
            file_content = filename.read_text(encoding='utf-8')
        except UnicodeDecodeError:
            session.log(f'File {filename.name} is not a valid UTF-8 file.')
            is_error = True
            continue

        # Check whether the content of the file is anyhow valid.
        is_valid, error_message = is_changelog_entry_valid(file_content)
        if not is_valid:
            session.log(f'File {filename.name} is not a valid changelog entry: {error_message}')
            is_error = True
            continue

    if is_error:
        session.error(
            'Found errors in the changelog.d directory. Check logs above for more information'
        )



######################################################################
#
# File: b2/_internal/version_listing.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import pathlib
import re
from typing import List

RE_VERSION = re.compile(r'[_]*b2v(\d+)')


def get_versions() -> List[str]:
    return [path.name for path in sorted(pathlib.Path(__file__).parent.glob('*b2v*'))]


def get_int_version(version: str) -> int:
    match = RE_VERSION.match(version)
    assert match, f'Version {version} does not match pattern {RE_VERSION.pattern}'
    return int(match.group(1))


CLI_VERSIONS = get_versions()
UNSTABLE_CLI_VERSION = max(CLI_VERSIONS, key=get_int_version)
LATEST_STABLE_VERSION = max(
    [elem for elem in CLI_VERSIONS if not elem.startswith('_')], key=get_int_version
)



######################################################################
#
# File: b2/_internal/_cli/__init__.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
"""
_cli package contains internals of the command-line interface to the B2.

It is not intended to be used as a library.
"""



######################################################################
#
# File: doc/source/conf.py
#
# Copyright 2019, Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# -*- coding: utf-8 -*-
#
# B2_Command_Line_Tool documentation build configuration file, created by
# sphinx-quickstart on Fri Oct 20 18:27:20 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import datetime
import importlib
import os
import re
import sys
import textwrap
from os import path

sys.path.insert(0, os.path.abspath('../..'))

from b2._internal.version import VERSION
from b2._internal.version_listing import LATEST_STABLE_VERSION

B2 = importlib.import_module(f'b2._internal.{LATEST_STABLE_VERSION}.registry').B2

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.intersphinx', 'sphinx.ext.ifconfig', 'sphinx.ext.viewcode',
    'sphinx.ext.coverage', 'sphinxarg.ext'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'B2_Command_Line_Tool'

year = datetime.date.today().strftime("%Y")
author = 'Backblaze'
copyright = f'{year}, {author}'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = VERSION.rsplit('.', 1)[0]
# The full version, including alpha/beta/rc tags.
release = VERSION

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = os.environ.get('B2_SPHINX_TODO', False) and True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'prev_next_buttons_location': 'both',
    'collapse_navigation': True,
}

autodoc_default_options = {
    'member-order': 'bysource',
    'exclude-members': '__weakref__, _abc_cache, _abc_negative_cache, _abc_negative_cache_version, _abc_registry, _abc_impl',
    'members': True,
    'undoc-members': True,
}  # yapf: disable

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**':
        [
            'about.html',
            'navigation.html',
            'relations.html',  # needs 'show_related': True theme option to display
            'searchbox.html',
            'donate.html',
        ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'B2_Command_Line_Tooldoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc, 'B2_Command_Line_Tool.tex', 'B2\\_Command\\_Line\\_Tool Documentation',
        'Backblaze', 'manual'
    ),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'b2_command_line_tool', 'B2_Command_Line_Tool Documentation', [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc, 'B2_Command_Line_Tool', 'B2_Command_Line_Tool Documentation', author,
        'B2_Command_Line_Tool', 'One line description of project.', 'Miscellaneous'
    ),
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python3': ('https://docs.python.org/3', None)}

white_spaces_start = re.compile(r'^\s*')


def setup(_):
    """
    Building docs requires dumping the main CLI help text to an .rst file before the actual build. This is the most
    reasonable way to piggy back that behaviour. Checking if the new file contents would the same as the old one
    (if any) is important, so that the automatic file-watcher/doc-builder doesn't fall into an endless loop.
    """
    main_help_text = str(B2.lazy_get_description(NAME='b2'))
    main_help_text = textwrap.dedent(main_help_text)

    main_help_path = path.join(path.dirname(__file__), 'main_help.rst')
    if path.exists(main_help_path):
        with open(main_help_path) as main_help_file:
            if main_help_file.read() == main_help_text:
                return
    with open(main_help_path, 'w') as main_help_file:
        main_help_file.write(main_help_text)



######################################################################
#
# File: test/integration/test_tqdm_closer.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import re
import sys

import pytest


@pytest.mark.skipif(
    (sys.platform != 'darwin') or ((sys.version_info.major, sys.version_info.minor) < (3, 11)),
    reason='Tqdm closing error only occurs on OSX and python 3.11 or newer',
)
def test_tqdm_closer(b2_tool, bucket, file_name):
    # test that stderr doesn't contain any warning, in particular warnings about multiprocessing resource tracker
    # leaking semaphores
    b2_tool.should_succeed([
        'cat',
        f'b2://{bucket.name}/{file_name}',
    ])

    # test that disabling _TqdmCloser does produce a resource tracker warning. Should the following check ever fail,
    # that would mean that either Tqdm or python fixed the issue and _TqdmCloser can be disabled for fixed versions
    b2_tool.should_succeed(
        [
            'cat',
            f'b2://{bucket.name}/{file_name}',
        ],
        additional_env={'B2_TEST_DISABLE_TQDM_CLOSER': '1'},
        expected_stderr_pattern=re.compile(
            r'UserWarning: resource_tracker: There appear to be \d+ leaked semaphore'
            r' objects to clean up at shutdown'
        ),
    )



######################################################################
#
# File: b2/_internal/arg_parser.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import argparse
import functools
import locale
import re
import sys
import textwrap
import unittest.mock

from rst2ansi import rst2ansi


class B2RawTextHelpFormatter(argparse.RawTextHelpFormatter):
    """
    CLI custom formatter.

    It removes default "usage: " text and prints usage for all (non-hidden) subcommands.
    """

    def __init__(self, *args, show_all: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.show_all = show_all

    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = ''
        super().add_usage(usage, actions, groups, prefix)

    def add_argument(self, action):
        if isinstance(action, argparse._SubParsersAction) and action.help is not argparse.SUPPRESS:
            usages = []
            for choice in self._unique_choice_values(action):
                deprecated = getattr(choice, 'deprecated', False)
                if deprecated:
                    if self.show_all:
                        usages.append(f'(DEPRECATED) {choice.format_usage()}')
                else:
                    usages.append(choice.format_usage())
            self.add_text(''.join(usages))
        else:
            super().add_argument(action)

    @classmethod
    def _unique_choice_values(cls, action):
        seen = set()
        seen_add = seen.add
        for value in action.choices.values():
            if not (value in seen or seen_add(value)):
                yield value


class _HelpAllAction(argparse._HelpAction):
    """Like argparse._HelpAction but prints help for all subcommands (even deprecated ones)."""

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help(show_all=True)
        parser.exit()


class B2ArgumentParser(argparse.ArgumentParser):
    """
    CLI custom parser.

    It fixes indentation of the description, set the custom formatter as a default
    and use help message in case of error.
    """

    def __init__(
        self,
        *args,
        add_help_all: bool = True,
        for_docs: bool = False,
        deprecated: bool = False,
        **kwargs
    ):
        """

        :param for_docs: is this parser used for generating docs
        :param deprecated: is this option deprecated
        """
        self._raw_description = None
        self._description = None
        self._for_docs = for_docs
        self.deprecated = deprecated
        kwargs.setdefault('formatter_class', B2RawTextHelpFormatter)
        super().__init__(*args, **kwargs)
        if add_help_all:
            self.register('action', 'help_all', _HelpAllAction)
            self.add_argument(
                '--help-all',
                help='show help for all options, including deprecated ones',
                action='help_all',
            )

    @property
    def description(self):
        if self._description is None and self._raw_description is not None:
            if self._for_docs:
                self._description = textwrap.dedent(self._raw_description)
            else:
                encoding = self._get_encoding()
                self._description = rst2ansi(
                    self._raw_description.encode(encoding), output_encoding=encoding
                )

        return self._description

    @description.setter
    def description(self, value):
        self._raw_description = value

    def error(self, message):
        self.print_help()

        self.exit(2, f'\n{self.prog}: error: {message}\n')

    @classmethod
    def _get_encoding(cls):
        _, locale_encoding = locale.getdefaultlocale()

        # Check if the stdout is properly set
        if sys.stdout.encoding is not None:
            # Use the stdout encoding
            return sys.stdout.encoding

        # Fall back to the locale_encoding if stdout encoding is not set
        elif locale_encoding is not None:
            return locale_encoding

        # locales are improperly configured
        return 'ascii'

    def print_help(self, *args, show_all: bool = False, **kwargs):
        """
        Print help message.
        """
        with unittest.mock.patch.object(
            self, 'formatter_class', functools.partial(B2RawTextHelpFormatter, show_all=show_all)
        ):
            super().print_help(*args, **kwargs)


SUPPORT_CAMEL_CASE_ARGUMENTS = False


def enable_camel_case_arguments():
    global SUPPORT_CAMEL_CASE_ARGUMENTS
    SUPPORT_CAMEL_CASE_ARGUMENTS = True


def make_deprecated_action_call(action):
    def deprecated_action_call(self, parser, namespace, values, option_string=None, **kwargs):
        action.__call__(self, parser, namespace, values, option_string, **kwargs)
        if option_string:
            kebab_option_string = _camel_to_kebab(option_string)
            print(
                f"The '{option_string}' argument is deprecated. Use '{kebab_option_string}' instead.",
                file=sys.stderr
            )

    return deprecated_action_call


_kebab_to_snake_pattern = re.compile(r'-')
_camel_to_kebab_pattern = re.compile(r'(?<=[a-z])([A-Z])')
_kebab_to_camel_pattern = re.compile(r'-(\w)')


def _camel_to_kebab(s: str):
    return _camel_to_kebab_pattern.sub(r'-\1', s).lower()


def _kebab_to_camel(s: str):
    return "--" + _kebab_to_camel_pattern.sub(lambda m: m.group(1).upper(), s[2:])


def _kebab_to_snake(s: str):
    return _kebab_to_snake_pattern.sub('_', s)


class DeprecatedActionMarker:
    pass


def add_normalized_argument(parser, param_name, *args, **kwargs):
    param_name_kebab = _camel_to_kebab(param_name)
    param_name_camel = _kebab_to_camel(param_name_kebab)
    dest_name_snake = _kebab_to_snake(param_name_kebab)[2:]
    kwargs_kebab = dict(kwargs)
    kwargs_camel = kwargs
    kwargs_camel['help'] = argparse.SUPPRESS

    if 'dest' not in kwargs_kebab:
        kwargs_kebab['dest'] = dest_name_snake
        kwargs_camel['dest'] = dest_name_snake

    if 'action' in kwargs:
        if isinstance(kwargs['action'], str):
            action = parser._registry_get('action', kwargs['action'])
        else:
            action = kwargs['action']
    else:
        action = argparse._StoreAction

    kwargs_camel['action'] = type(
        'DeprecatedAction', (action, DeprecatedActionMarker),
        {'__call__': make_deprecated_action_call(action)}
    )

    parser.add_argument(f'{param_name_kebab}', *args, **kwargs_kebab)

    if SUPPORT_CAMEL_CASE_ARGUMENTS and param_name_kebab != param_name_camel:
        parser.add_argument(f'{param_name_camel}', *args, **kwargs_camel)



######################################################################
#
# File: test/unit/test_arg_parser.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

import argparse
import sys

from b2._internal._cli.arg_parser_types import (
    parse_comma_separated_list,
    parse_millis_from_float_timestamp,
    parse_range,
)
from b2._internal.arg_parser import B2ArgumentParser
from b2._internal.console_tool import B2

from .test_base import TestBase


class TestCustomArgTypes(TestBase):
    def test_parse_comma_separated_list(self):
        self.assertEqual([''], parse_comma_separated_list(''))
        self.assertEqual(['1', '2', '3'], parse_comma_separated_list('1,2,3'))

    def test_parse_millis_from_float_timestamp(self):
        self.assertEqual(1367900664000, parse_millis_from_float_timestamp('1367900664'))
        self.assertEqual(1367900664152, parse_millis_from_float_timestamp('1367900664.152'))
        with self.assertRaises(ValueError):
            parse_millis_from_float_timestamp('!$@$%@!@$')

    def test_parse_range(self):
        self.assertEqual((1, 2), parse_range('1,2'))
        with self.assertRaises(argparse.ArgumentTypeError):
            parse_range('1')
        with self.assertRaises(argparse.ArgumentTypeError):
            parse_range('1,2,3')
        with self.assertRaises(ValueError):
            parse_range('!@#,%^&')


class TestNonUTF8TerminalSupport(TestBase):
    class ASCIIEncodedStream:
        def __init__(self, original_stream):
            self.original_stream = original_stream
            self.encoding = 'ascii'

        def write(self, data):
            if isinstance(data, str):
                data = data.encode(self.encoding, 'strict')
            self.original_stream.buffer.write(data)

        def flush(self):
            self.original_stream.flush()

    def check_help_string(self, command_class, command_name):
        help_string = command_class.__doc__

        # create a parser with a help message that is based on the command_class.__doc__ string
        parser = B2ArgumentParser(description=help_string)

        try:
            old_stdout = sys.stdout
            old_stderr = sys.stderr
            sys.stdout = TestNonUTF8TerminalSupport.ASCIIEncodedStream(sys.stdout)
            sys.stderr = TestNonUTF8TerminalSupport.ASCIIEncodedStream(sys.stderr)

            parser.print_help()

        except UnicodeEncodeError as e:
            self.fail(
                f'Failed to encode help message for command "{command_name}" on a non-UTF-8 terminal: {e}'
            )

        finally:
            # Restore original stdout and stderr
            sys.stdout = old_stdout
            sys.stderr = old_stderr

    def test_help_in_non_utf8_terminal(self):
        command_classes = dict(B2.subcommands_registry.items())
        command_classes['b2'] = B2

        for command_name, command_class in command_classes.items():
            with self.subTest(command_class=command_class, command_name=command_name):
                self.check_help_string(command_class, command_name)
